from smc.multiprocessing.networking.client import client_receiver, client_sender
from smc.multiprocessing.process_manager import ProcessManager
from smc.robots.interfaces.force_torque_sensor_interface import (
    ForceTorqueOnSingleArmWrist,
)
from smc.control.cartesian_space.ik_solvers import getIKSolver
from smc.control.cartesian_space.cartesian_space_point_to_point import moveL
from smc.control.cartesian_space.utils import getClikArgs
from smc.control.control_loop_manager import ControlLoopManager
from smc import (
    getMinimalArgParser,
    getRobotFromArgs,
)

import pinocchio as pin
import numpy as np
from functools import partial


def get_args():
    parser = getMinimalArgParser()
    parser = getClikArgs(parser)
    parser.description = "the robot will received joint angles from a socket and go to them in joint space"
    # add more arguments here from different Simple Manipulator Control modules
    parser.add_argument("--host", type=str, help="host ip address", default="127.0.0.1")
    parser.add_argument("--port-wrench", type=int, help="host's port", default=6666)
    parser.add_argument("--port-command", type=int, help="host's port", default=7777)
    # put your own default here
    # it's translation + orientation as a quarternion
    parser.add_argument(
        "--init-pose", type=float, nargs=7, default=[0.3, 0.3, 0.3, 0.0, 0.0, 0.0, 1.0]
    )
    args = parser.parse_args()
    return args


def controlLoopClikExternalGoal(
    robot: ForceTorqueOnSingleArmWrist,
    receiver: ProcessManager,
    sender: ProcessManager,
    ik_solver,
    i,
    past_data,
) -> tuple[bool, dict[str, np.ndarray], dict[str, np.ndarray]]:
    """
    controlLoop
    -----------------------------
    controller description
    """
    breakFlag = False
    log_item = {}
    save_past_dict = {}

    q = robot.q
    T_w_e = robot.T_w_e
    data = receiver.getData()
    T_goal = data["T_goal"]
    # print(T_goal)
    # print("translation:", T_goal.translation)
    # print("rotation:", pin.Quaternion(T_goal.rotation))
    SEerror = T_w_e.actInv(T_goal)
    err_vector = pin.log6(SEerror).vector
    J = pin.computeFrameJacobian(robot.model, robot.data, q, robot.ee_frame_id)
    K = np.diag([30] * 3 + [15] * 3)

    qd_cmd = (
        J.T
        @ np.linalg.inv(J @ J.T + np.eye(J.shape[0], J.shape[0]) * args.tikhonov_damp)
        @ (K @ err_vector + data["v"])
    )
    if qd_cmd is None:
        print("the controller you chose didn't work, using dampedPseudoinverse instead")
        qd_cmd = ik_solver(1e-2, J, err_vector)

    # NOTE: this will also receive a velocity
    # i have no idea what to do with that velocity
    robot.sendVelocityCommand(qd_cmd)
    # NOTE: this one is in the base frame! (as it should be)
    # it is saved in robotmanager (because it's a private variable,
    # yes this is evil bla bla)
    # NOTE: idk whether i'm sending this or wrench_ee lule
    sender.sendCommand({"wrench": robot.wrench_base})

    log_item["qs"] = robot.q
    log_item["dqs"] = robot.v
    log_item["err_vector"] = err_vector
    return breakFlag, save_past_dict, log_item


if __name__ == "__main__":
    args = get_args()
    robot = getRobotFromArgs(args)
    translation = np.array(args.init_pose[:3])
    orientation = pin.Quaternion(np.array(args.init_pose[3:]))
    init_goal = pin.SE3(orientation, translation)
    moveL(args, robot, init_goal)
    print("arrived at initial position")

    # VERY important that the first q we'll pass as desired is the current q, meaning the robot won't move
    # this is set with init_value
    # command_sender: 7777
    args.port = args.port_command
    receiver = ProcessManager(
        args,
        client_receiver,
        {"T_goal": pin.SE3.Identity(), "v": np.zeros(6)},
        4,
        init_value={"T_goal": robot.T_w_e, "v": np.zeros(6)},
    )
    # wrench_sender: 6666
    args.port = args.port_wrench
    sender = ProcessManager(args, client_sender, {"wrench": np.zeros(6)}, 0)
    log_item = {
        "qs": np.zeros((robot.model.nq,)),
        "dqs": np.zeros((robot.model.nv,)),
        "err_vector": np.zeros((6,)),
    }
    clik_controller = getIKSolver(args, robot)
    control_loop = partial(
        controlLoopClikExternalGoal, robot, receiver, sender, clik_controller
    )
    loop_manager = ControlLoopManager(robot, control_loop, args, {}, log_item)
    loop_manager.run()

    # get expected behaviour here (library can't know what the end is - you have to do this here)
    if args.real:
        robot.stopRobot()

    if args.save_log:
        robot._log_manager.saveLog()
        robot._log_manager.plotAllControlLoops()

    if args.visualizer:
        robot.killManipulatorVisualizer()
